import logging

from typing import Any, Optional

from polyaxon._client.decorators import client_handler
from polyaxon._schemas.lifecycle import V1StatusCondition, V1Statuses
from traceml import tracking
from traceml.events.schemas import V1EventSpan, V1EventSpanKind
from traceml.exceptions import TracemlException

try:
    from langchain.callbacks.tracers.base import BaseTracer
    from langchain.callbacks.tracers.schemas import Run, TracerSession
    from langchain.env import get_runtime_environment
    from langchain.load.dump import dumpd
    from langchain.schema.messages import BaseMessage
except ImportError:
    raise TracemlException("Langchain is required to use the tracking Callback")

logger = logging.getLogger(__name__)


def _serialize_io(run_inputs: dict) -> dict:
    serialized_inputs = {}
    for key, value in run_inputs.items():
        if key == "input_documents":
            serialized_inputs.update(
                {f"input_document_{i}": doc.to_json() for i, doc in enumerate(value)}
            )
        else:
            serialized_inputs[key] = value
    return serialized_inputs


class RunProcessor:
    """Handles the conversion of a LangChain Runs into a trace."""

    @classmethod
    def process_span(cls, run: Run) -> Optional["V1EventSpan"]:
        """Converts a LangChain Run into a V1EventSpan.
        Params:
            run: The LangChain Run to convert.
        Returns:
            The converted V1EventSpan.
        """
        try:
            span = cls._convert_lc_run_to_span(run)
            return span
        except Exception as e:
            logger.warning(
                f"Skipping trace saving - unable to safely convert LangChain Run "
                f"into Trace due to: {e}"
            )
            return None

    @classmethod
    def _convert_run_to_span(cls, run: Run) -> "V1EventSpan":
        """Base utility to create a span from a run.

        Params:
            run: The run to convert.
        Returns:
            The converted V1EventSpan.
        """
        metadata = {**run.extra} if run.extra else {}
        metadata["execution_order"] = run.execution_order

        status_conditions = (
            [
                V1StatusCondition.construct(
                    type=V1Statuses.FAILED,
                    status=True,
                    reason="SpanFailed",
                    message=run.error,
                    last_transition_time=run.end_time,
                    last_update_time=run.end_time,
                )
            ]
            if run.error
            else None
        )
        return V1EventSpan(
            uuid=str(run.id) if run.id is not None else None,
            name=run.name,
            started_at=run.start_time,
            finished_at=run.end_time,
            status=V1Statuses.SUCCEEDED if run.error is None else V1Statuses.FAILED,
            status_conditions=status_conditions,
            metadata=metadata,
        )

    @classmethod
    def _convert_llm_run_to_span(cls, run: Run) -> "V1EventSpan":
        """Converts a LangChain LLM Run into a V1EventSpan.
        Params
            run: The LangChain LLM Run to convert.
        Returns:
            The converted V1EventSpan.
        """
        base_span = cls._convert_run_to_span(run)
        if base_span.metadata is None:
            base_span.metadata = {}
        base_span.inputs = run.inputs
        base_span.outputs = run.outputs
        base_span.metadata["llm_output"] = run.outputs.get("llm_output", {})
        base_span.kind = V1EventSpanKind.LLM
        return base_span

    @classmethod
    def _convert_chain_run_to_span(cls, run: Run) -> "V1EventSpan":
        """Converts a LangChain Chain Run into a V1EventSpan.

        Params
            run: The LangChain Chain Run to convert.
        Returns:
            The converted V1EventSpan.
        """
        base_span = cls._convert_run_to_span(run)

        base_span.inputs = _serialize_io(run.inputs)
        base_span.outputs = _serialize_io(run.outputs)
        base_span.children = [
            cls._convert_lc_run_to_span(child_run) for child_run in run.child_runs
        ]
        base_span.kind = (
            V1EventSpanKind.AGENT
            if "agent" in run.name.lower()
            else V1EventSpanKind.CHAIN
        )

        return base_span

    @classmethod
    def _convert_tool_run_to_span(cls, run: Run) -> "V1EventSpan":
        """Converts a LangChain Tool Run into a V1EventSpan.

        Params
            run: The LangChain Tool Run to convert.
        Returns:
            The converted V1EventSpan.
        """
        base_span = cls._convert_run_to_span(run)
        base_span.inputs = _serialize_io(run.inputs)
        base_span.outputs = _serialize_io(run.outputs)
        base_span.children = [
            cls._convert_lc_run_to_span(child_run) for child_run in run.child_runs
        ]
        base_span.kind = V1EventSpanKind.TOOL

        return base_span

    @classmethod
    def _convert_lc_run_to_span(cls, run: Run) -> "V1EventSpan":
        """Utility to convert any generic LangChain Run into a V1EventSpan.

        Params
            run: The LangChain Run to convert.
        Returns:
            The converted V1EventSpan.
        """
        if run.run_type == V1EventSpanKind.LLM:
            return cls._convert_llm_run_to_span(run)
        elif run.run_type == V1EventSpanKind.CHAIN:
            return cls._convert_chain_run_to_span(run)
        elif run.run_type == V1EventSpanKind.TOOL:
            return cls._convert_tool_run_to_span(run)
        else:
            return cls._convert_run_to_span(run)


class Callback(BaseTracer):
    """Callback Handler that logs Langchain traces/spans."""

    @client_handler(check_no_op=True)
    def __init__(self, run=None, **kwargs: Any) -> None:
        """Initializes the callback.

        To monitor all LangChain activity, add this tracer like any other
        LangChain callback:

        ```python
        from polyaxon.tracking.integrations.langchain import Callback
        callback = Callback()
        chain = LLMChain(llm, callbacks=[callback])
        ```

        When using manual tracking of multiple runs in a single script:

        ```python
        from polyaxon.tracking.integrations.langchain import Callback
        tracking.init(..., is_new=True, ...)
        ...
        callback = Callback()
        chain = LLMChain(llm, callbacks=[callback])
        ...
        tracking.end()
        ```
        """
        super().__init__(**kwargs)
        self.run = tracking.get_or_create_run(run)

    @client_handler(check_no_op=True)
    def _persist_run(self, run: "Run"):
        """Converts a LangChain Run to a Trace."""
        span = RunProcessor.process_span(run)
        if span is None:
            return
        if self.run is not None:
            self.run.log_trace(span=span)



import os

from typing import List

from clipped.utils.enums import get_enum_value
from clipped.utils.paths import check_or_create_path, set_permissions

from traceml.artifacts import V1ArtifactKind
from traceml.events import (
    LoggedEventListSpec,
    LoggedEventSpec,
    get_event_path,
    get_resource_path,
)


class EventWriter:
    EVENTS_BACKEND = "events"
    RESOURCES_BACKEND = "resources"

    def __init__(self, run_path: str, backend: str):
        self._events_backend = backend
        self._run_path = run_path
        self._files = {}  # type: dict[str, LoggedEventListSpec]
        self._closed = False

    def _get_event_path(self, kind: str, name: str) -> str:
        if self._events_backend == self.EVENTS_BACKEND:
            return get_event_path(
                run_path=self._run_path,
                kind=kind,
                name=name,
            )
        if self._events_backend == self.RESOURCES_BACKEND:
            return get_resource_path(
                run_path=self._run_path,
                kind=kind,
                name=name,
            )
        raise ValueError(
            "Unrecognized backend {}".format(get_enum_value(self._events_backend))
        )

    def _init_events(self, events_spec: LoggedEventListSpec):
        event_path = self._get_event_path(kind=events_spec.kind, name=events_spec.name)
        # Check if the file exists otherwise initialize
        if not os.path.exists(event_path):
            check_or_create_path(event_path, is_dir=False)
            with open(event_path, "w") as event_file:
                if V1ArtifactKind.is_jsonl_file_event(events_spec.kind):
                    event_file.write("")
                else:
                    event_file.write(events_spec.get_csv_header())
            set_permissions(event_path)

    def _append_events(self, events_spec: LoggedEventListSpec):
        event_path = self._get_event_path(kind=events_spec.kind, name=events_spec.name)
        with open(event_path, "a") as event_file:
            if V1ArtifactKind.is_jsonl_file_event(events_spec.kind):
                event_file.write(events_spec.get_jsonl_events())
            else:
                event_file.write(events_spec.get_csv_events())

    def _events_to_files(self, events: List[LoggedEventSpec]):
        for event in events:
            file_name = "{}.{}".format(event.kind, event.name)
            if file_name in self._files:
                self._files[file_name].events.append(event.event)
            else:
                self._files[file_name] = LoggedEventListSpec(
                    kind=event.kind, name=event.name, events=[event.event]
                )
                self._init_events(self._files[file_name])

    def write(self, events: List[LoggedEventSpec]):
        if not events:
            return
        if isinstance(events, LoggedEventSpec):
            events = [events]
        self._events_to_files(events)

    def flush(self):
        for file_name in self._files:
            events_spec = self._files[file_name]
            if events_spec.events:
                self._append_events(events_spec)
            self._files[file_name].empty_events()

    def close(self):
        self.flush()
        self._closed = True

    @property
    def closed(self):
        return self._closed


class BaseFileWriter:
    """Writes `LoggedEventSpec` to event files.

    The `EventFileWriter` class creates a event files in the run path,
    and asynchronously writes Events to the files.
    """

    def __init__(self, run_path: str):
        self._run_path = run_path
        check_or_create_path(run_path, is_dir=True)

    @property
    def run_path(self):
        return self._run_path

    def add_event(self, event: LoggedEventSpec):
        if not isinstance(event, LoggedEventSpec):
            raise TypeError("Expected an LoggedEventSpec, " " but got %s" % type(event))
        self._async_writer.write(event)

    def add_events(self, events: List[LoggedEventSpec]):
        for e in events:
            if not isinstance(e, LoggedEventSpec):
                raise TypeError("Expected an LoggedEventSpec, " " but got %s" % type(e))
        self._async_writer.write(events)

    def flush(self):
        """Flushes the event files to disk.

        Call this method to make sure that all pending events have been
        written to disk.
        """
        self._async_writer.flush()

    def close(self):
        """Performs a final flush of the event files to disk, stops the
        write/flush worker and closes the files.

        Call this method when you do not need the writer anymore.
        """
        self._async_writer.close()



NAME = "traceml"
VERSION = "1.2.1-rc2"
DESC = (
    "Engine for ML/Data tracking, visualization, dashboards, and model UI for Polyaxon."
)
URL = "https://github.com/polyaxon/traceml"
AUTHOR = "Polyaxon, Inc."
EMAIL = "contact@polyaxon.com"
LICENSE = "Apache 2.0"



from typing import Optional

from clipped.utils.paths import copy_file_or_dir_path

from traceml.events import V1EventArtifact

try:
    import numpy as np
except ImportError:
    np = None


def artifact_path(
    from_path: str, asset_path: str, kind: str, asset_rel_path: Optional[str] = None
) -> V1EventArtifact:
    copy_file_or_dir_path(from_path, asset_path)
    return V1EventArtifact(kind=kind, path=asset_rel_path or asset_path)



from typing import Optional, Union

from clipped.utils.enums import PEnum


class V1ArtifactKind(str, PEnum):
    MODEL = "model"
    AUDIO = "audio"
    VIDEO = "video"
    HISTOGRAM = "histogram"
    IMAGE = "image"
    TENSOR = "tensor"
    DATAFRAME = "dataframe"
    CHART = "chart"
    CSV = "csv"
    TSV = "tsv"
    PSV = "psv"
    SSV = "ssv"
    METRIC = "metric"
    ENV = "env"
    HTML = "html"
    TEXT = "text"
    FILE = "file"
    DIR = "dir"
    DOCKERFILE = "dockerfile"
    DOCKER_IMAGE = "docker_image"
    DATA = "data"
    CODEREF = "coderef"
    TABLE = "table"
    TENSORBOARD = "tensorboard"
    CURVE = "curve"
    CONFUSION = "confusion"
    ANALYSIS = "analysis"
    ITERATION = "iteration"
    MARKDOWN = "markdown"
    SYSTEM = "system"
    ARTIFACT = "artifact"
    SPAN = "span"

    @classmethod
    def is_jsonl_file_event(cls, kind: Optional[Union["V1ArtifactKind", str]]) -> bool:
        return kind in {
            V1ArtifactKind.HTML,
            V1ArtifactKind.TEXT,
            V1ArtifactKind.CHART,
            V1ArtifactKind.SPAN,
        }

    @classmethod
    def is_single_file_event(cls, kind: Optional[Union["V1ArtifactKind", str]]) -> bool:
        return kind in {
            V1ArtifactKind.HTML,
            V1ArtifactKind.TEXT,
            V1ArtifactKind.HISTOGRAM,
            V1ArtifactKind.CHART,
            V1ArtifactKind.CONFUSION,
            V1ArtifactKind.CURVE,
            V1ArtifactKind.METRIC,
            V1ArtifactKind.SYSTEM,
            V1ArtifactKind.SPAN,
        }

    @classmethod
    def is_single_or_multi_file_event(
        cls, kind: Optional[Union["V1ArtifactKind", str]]
    ) -> bool:
        return kind in {
            V1ArtifactKind.MODEL,
            V1ArtifactKind.DATAFRAME,
            V1ArtifactKind.AUDIO,
            V1ArtifactKind.VIDEO,
            V1ArtifactKind.IMAGE,
            V1ArtifactKind.CSV,
            V1ArtifactKind.TSV,
            V1ArtifactKind.PSV,
            V1ArtifactKind.SSV,
        }

    @classmethod
    def is_dir(cls, kind: Optional[Union["V1ArtifactKind", str]]) -> bool:
        return kind in {
            V1ArtifactKind.TENSORBOARD,
            V1ArtifactKind.DIR,
        }

    @classmethod
    def is_file(cls, kind: Optional[Union["V1ArtifactKind", str]]) -> bool:
        return kind in {
            V1ArtifactKind.DOCKERFILE,
            V1ArtifactKind.FILE,
            V1ArtifactKind.ENV,
        }

    @classmethod
    def is_file_or_dir(cls, kind: Optional[Union["V1ArtifactKind", str]]) -> bool:
        return kind in {
            V1ArtifactKind.DATA,
            V1ArtifactKind.MODEL,
        }



from typing import Dict, List

from traceml.processors.events_processors import metrics_dict_to_list

try:
    import psutil
except ImportError:
    psutil = None


def can_log_psutil_resources():
    return psutil is not None


def query_psutil() -> Dict:
    results = {}
    try:
        # psutil <= 5.6.2 did not have getloadavg:
        if hasattr(psutil, "getloadavg"):
            results["load"] = psutil.getloadavg()[0]
        else:
            # Do not log an empty metric
            pass
    except OSError:
        pass
    vm = psutil.virtual_memory()
    results["cpu"] = psutil.cpu_percent(interval=None)
    results["memory"] = vm.percent
    return results


def get_psutils_metrics() -> List:
    return metrics_dict_to_list(query_psutil())



import io

from typing import Optional

from clipped.utils.np import calculate_scale_factor, to_np
from clipped.utils.paths import check_or_create_path, copy_file_path

from polyaxon._constants.globals import UNKNOWN
from traceml.events import V1EventImage
from traceml.logger import logger
from traceml.processors.errors import (
    MATPLOTLIB_ERROR_MESSAGE,
    NUMPY_ERROR_MESSAGE,
    PIL_ERROR_MESSAGE,
)

try:
    import numpy as np
except ImportError:
    np = None


def image_path(
    from_path: str, asset_path: str, asset_rel_path: Optional[str] = None
) -> V1EventImage:
    copy_file_path(from_path, asset_path)
    return V1EventImage(path=asset_rel_path or asset_path)


def _draw_single_box(
    image,
    xmin,
    ymin,
    xmax,
    ymax,
    display_str,
    color="black",
    color_text="black",
    thickness=2,
):
    if not np:
        logger.warning(NUMPY_ERROR_MESSAGE)
        return UNKNOWN

    try:
        from PIL import ImageDraw, ImageFont
    except ImportError:
        logger.warning(PIL_ERROR_MESSAGE)
        return UNKNOWN

    font = ImageFont.load_default()
    draw = ImageDraw.Draw(image)
    (left, right, top, bottom) = (xmin, xmax, ymin, ymax)
    draw.line(
        [(left, top), (left, bottom), (right, bottom), (right, top), (left, top)],
        width=thickness,
        fill=color,
    )
    if display_str:
        text_bottom = bottom
        # Reverse list and print from bottom to top.
        text_width, text_height = font.getsize(display_str)
        margin = np.ceil(0.05 * text_height)
        draw.rectangle(
            [
                (left, text_bottom - text_height - 2 * margin),
                (left + text_width, text_bottom),
            ],
            fill=color,
        )
        draw.text(
            (left + margin, text_bottom - text_height - margin),
            display_str,
            fill=color_text,
            font=font,
        )
    return image


def encoded_image(asset_path: str, data, asset_rel_path: Optional[str] = None):
    try:
        from PIL import Image
    except ImportError:
        logger.warning(PIL_ERROR_MESSAGE)
        return UNKNOWN

    image_data = Image.open(io.BytesIO(data.encoded_image_string))
    return save_image(
        asset_path=asset_rel_path or asset_path,
        image_data=image_data,
        height=data.height,
        width=data.width,
        colorspace=data.colorspace,
    )


def image(
    asset_path: str,
    data,
    rescale=1,
    dataformats="CHW",
    asset_rel_path: Optional[str] = None,
):
    if not np:
        logger.warning(NUMPY_ERROR_MESSAGE)
        return UNKNOWN

    tensor = to_np(data)
    tensor = convert_to_HWC(tensor, dataformats)
    # Do not assume that user passes in values in [0, 255], use data type to detect
    scale_factor = calculate_scale_factor(tensor)
    tensor = tensor.astype(np.float32)
    tensor = (tensor * scale_factor).astype(np.uint8)
    return make_image(
        asset_path, tensor, rescale=rescale, asset_rel_path=asset_rel_path
    )


def image_boxes(
    asset_path: str,
    tensor_image,
    tensor_boxes,
    rescale=1,
    dataformats="CHW",
    asset_rel_path: Optional[str] = None,
):
    if not np:
        logger.warning(NUMPY_ERROR_MESSAGE)
        return UNKNOWN

    tensor_image = to_np(tensor_image)
    tensor_image = convert_to_HWC(tensor_image, dataformats)
    tensor_boxes = to_np(tensor_boxes)
    tensor_image = tensor_image.astype(np.float32) * calculate_scale_factor(
        tensor_image
    )
    return make_image(
        asset_path,
        tensor_image.astype(np.uint8),
        rescale=rescale,
        rois=tensor_boxes,
        asset_rel_path=asset_rel_path,
    )


def draw_boxes(disp_image, boxes):
    # xyxy format
    num_boxes = boxes.shape[0]
    list_gt = range(num_boxes)
    for i in list_gt:
        disp_image = _draw_single_box(
            disp_image,
            boxes[i, 0],
            boxes[i, 1],
            boxes[i, 2],
            boxes[i, 3],
            display_str=None,
            color="Red",
        )
    return disp_image


def make_image(
    asset_path: str, tensor, rescale=1, rois=None, asset_rel_path: Optional[str] = None
):
    try:
        from PIL import Image
    except ImportError:
        logger.warning(PIL_ERROR_MESSAGE)
        return UNKNOWN

    height, width, colorspace = tensor.shape
    scaled_height = int(height * rescale)
    scaled_width = int(width * rescale)
    image_data = Image.fromarray(tensor)
    if rois is not None:
        image_data = draw_boxes(image_data, rois)
    image_data = image_data.resize((scaled_width, scaled_height), Image.LANCZOS)

    return save_image(
        asset_path=asset_path,
        image_data=image_data,
        height=height,
        width=width,
        colorspace=colorspace,
        asset_rel_path=asset_rel_path,
    )


def save_image(
    asset_path: str,
    image_data,
    height,
    width,
    colorspace,
    asset_rel_path: Optional[str] = None,
):
    check_or_create_path(asset_path, is_dir=False)
    image_data.save(asset_path, format="PNG")
    return V1EventImage(
        height=height,
        width=width,
        colorspace=colorspace,
        path=asset_rel_path or asset_path,
    )


def figure_to_image(figure, close: bool = True):
    """Render matplotlib figure to numpy format.

    Returns:
        numpy.array: image in [CHW] order
    """
    if not np:
        logger.warning(NUMPY_ERROR_MESSAGE)

    try:
        import matplotlib.backends.backend_agg as plt_backend_agg
        import matplotlib.pyplot as plt
    except ImportError:
        logger.warning(MATPLOTLIB_ERROR_MESSAGE)

    canvas = plt_backend_agg.FigureCanvasAgg(figure)
    canvas.draw()
    data = np.frombuffer(canvas.buffer_rgba(), dtype=np.uint8)
    w, h = figure.canvas.get_width_height()
    image_hwc = data.reshape([h, w, 4])[:, :, 0:3]
    image_chw = np.moveaxis(image_hwc, source=2, destination=0)
    if close:
        try:
            plt.close(figure.number)
        except Exception:  # noqa
            plt.close(figure)
    return image_chw


def figures_to_images(figures, close=True):
    """Render matplotlib figure to numpy format.

    Returns:
        numpy.array: image in [CHW] order
    """
    if not np:
        logger.warning(NUMPY_ERROR_MESSAGE)
        return UNKNOWN

    images = [figure_to_image(figure, close=close) for figure in figures]
    return np.stack(images)


def ensure_matplotlib_figure(figure):
    """Extract the current figure from a matplotlib object or return the object if it's a figure.
    raises ValueError if the object can't be converted.
    """
    try:
        import matplotlib

        from matplotlib.figure import Figure
    except ImportError:
        logger.warning(MATPLOTLIB_ERROR_MESSAGE)

    if figure == matplotlib.pyplot:
        figure = figure.gcf()
    elif not isinstance(figure, Figure):
        if hasattr(figure, "figure"):
            figure = figure.figure
            # Some matplotlib objects have a figure function
            if not isinstance(figure, Figure):
                raise ValueError(
                    "Only matplotlib.pyplot or matplotlib.pyplot.Figure objects are accepted."
                )
    if not figure.gca().has_data():
        raise ValueError(
            "You attempted to log an empty plot, "
            "pass a figure directly or ensure the global plot isn't closed."
        )
    return figure


def make_grid(data, ncols=8):
    # I: N1HW or N3HW
    if not np:
        logger.warning(NUMPY_ERROR_MESSAGE)
        return UNKNOWN

    assert isinstance(data, np.ndarray), "plugin error, should pass numpy array here"
    if data.shape[1] == 1:
        data = np.concatenate([data, data, data], 1)
    assert data.ndim == 4 and data.shape[1] == 3 or data.shape[1] == 4
    nimg = data.shape[0]
    H = data.shape[2]  # noqa
    W = data.shape[3]  # noqa
    ncols = min(nimg, ncols)
    nrows = int(np.ceil(float(nimg) / ncols))
    canvas = np.zeros((data.shape[1], H * nrows, W * ncols))
    i = 0
    for y in range(nrows):
        for x in range(ncols):
            if i >= nimg:
                break
            canvas[:, y * H : (y + 1) * H, x * W : (x + 1) * W] = data[i]  # noqa
            i = i + 1
    return canvas


def convert_to_HWC(tensor, input_format):  # noqa
    if not np:
        logger.warning(NUMPY_ERROR_MESSAGE)
        return UNKNOWN

    assert len(set(input_format)) == len(
        input_format
    ), "You can not use the same dimension shorthand twice. \
        input_format: {}".format(input_format)
    assert len(tensor.shape) == len(
        input_format
    ), "size of input tensor and input format are different. \
        tensor shape: {}, input_format: {}".format(tensor.shape, input_format)
    input_format = input_format.upper()

    if len(input_format) == 4:
        index = [input_format.find(c) for c in "NCHW"]
        tensor_NCHW = tensor.transpose(index)  # noqa
        tensor_CHW = make_grid(tensor_NCHW)  # noqa
        return tensor_CHW.transpose(1, 2, 0)

    if len(input_format) == 3:
        index = [input_format.find(c) for c in "HWC"]
        tensor_HWC = tensor.transpose(index)  # noqa
        if tensor_HWC.shape[2] == 1:
            tensor_HWC = np.concatenate([tensor_HWC, tensor_HWC, tensor_HWC], 2)  # noqa
        return tensor_HWC

    if len(input_format) == 2:
        index = [input_format.find(c) for c in "HW"]
        tensor = tensor.transpose(index)
        tensor = np.stack([tensor, tensor, tensor], 2)
        return tensor



from typing import Dict, Optional

from clipped.utils.paths import copy_file_or_dir_path

from traceml.events import V1EventModel
from traceml.logger import logger

try:
    import numpy as np
except ImportError:
    np = None


def model_path(
    from_path: str,
    asset_path: str,
    framework: Optional[str] = None,
    spec: Optional[Dict] = None,
    asset_rel_path: Optional[str] = None,
) -> V1EventModel:
    copy_file_or_dir_path(from_path, asset_path)
    return V1EventModel(
        path=asset_rel_path or asset_path, framework=framework, spec=spec
    )


def _model_to_str(model):
    filetype = "txt"
    if hasattr(model, "to_json"):
        model = model.model.to_json()
        filetype = "json"
    elif hasattr(model, "to_yaml"):
        model = model.to_yaml()
        filetype = "yaml"

    try:
        return str(model), filetype
    except Exception as e:
        logger.warning("Could not convert model to a string. Error: %s" % e)


def model_to_str(model):
    # Tensorflow Graph Definition
    if type(model).__name__ == "Graph":
        try:
            from google.protobuf import json_format

            graph_def = model.as_graph_def()
            model = json_format.MessageToJson(graph_def, sort_keys=True)
        except Exception as e:  # noqa
            logger.warning(
                "Could not convert Tensorflow graph to JSON %s", e, exc_info=True
            )

    return _model_to_str(model)



from typing import Dict, List

from traceml.logger import logger
from traceml.processors.events_processors import metrics_dict_to_list
from traceml.vendor import pynvml

try:
    import psutil
except ImportError:
    psutil = None


def can_log_gpu_resources():
    if pynvml is None:
        return False

    try:
        pynvml.nvmlInit()
        return True
    except pynvml.NVMLError:
        return False


def query_gpu(handle_idx: int, handle: any) -> Dict:
    memory = pynvml.nvmlDeviceGetMemoryInfo(handle)  # in Bytes
    utilization = pynvml.nvmlDeviceGetUtilizationRates(handle)

    return {
        "gpu_{}_memory_free".format(handle_idx): int(memory.free),
        "gpu_{}_memory_used".format(handle_idx): int(memory.used),
        "gpu_{}_utilization".format(handle_idx): utilization.gpu,
    }


def get_gpu_metrics() -> List:
    try:
        pynvml.nvmlInit()
        device_count = pynvml.nvmlDeviceGetCount()
        results = []

        for i in range(device_count):
            handle = pynvml.nvmlDeviceGetHandleByIndex(i)
            results += metrics_dict_to_list(query_gpu(i, handle))
        return results
    except pynvml.NVMLError:
        logger.debug("Failed to collect gpu resources", exc_info=True)
        return []



import logging
import sys

from traceml.logging.handler import LogStreamHandler, LogStreamWriter

EXCLUDE_DEFAULT_LOGGERS = ("polyaxon.client", "polyaxon.cli", "traceml")


def start_log_processor(add_logs, exclude=EXCLUDE_DEFAULT_LOGGERS):
    plx_logger = logging.getLogger("__plx__")
    plx_logger.setLevel(logging.INFO)
    if LogStreamHandler in map(type, plx_logger.handlers):
        for handler in plx_logger.handlers:
            if isinstance(handler, LogStreamHandler):
                handler.set_add_logs(add_logs=add_logs)
    else:
        handler = LogStreamHandler(add_logs=add_logs)
        plx_logger.addHandler(handler)

    exclude = ("__plx__",) + (exclude or ())
    for logger_name in exclude:
        logger = logging.getLogger(logger_name)
        if logging.StreamHandler not in map(type, logger.handlers):
            logger.addHandler(logging.StreamHandler())
            logger.propagate = False

    sys.stdout = LogStreamWriter(plx_logger, logging.INFO, sys.__stdout__)
    sys.stderr = LogStreamWriter(plx_logger, logging.ERROR, sys.__stderr__)


def end_log_processor():
    sys.stdout = sys.__stdout__
    sys.stderr = sys.__stderr__



from clipped.utils.np import sanitize_dict

from traceml import tracking
from traceml.exceptions import TracemlException

try:
    from sklearn.base import is_classifier, is_regressor
    from sklearn.metrics import (
        accuracy_score,
        explained_variance_score,
        f1_score,
        max_error,
        mean_absolute_error,
        precision_recall_fscore_support,
        r2_score,
    )
except ImportError:
    raise TracemlException("sklearn is required to use scikit polyaxon's loggers")


def _log_test_predictions(run, y_test, y_pred=None, nrows=1000):
    try:
        import pandas as pd
    except ImportError:
        return

    # single output
    if len(y_pred.shape) == 1:
        df = pd.DataFrame(data={"y_true": y_test, "y_pred": y_pred})
        run.log_dataframe(df=df.head(nrows), name="test_predictions")

    # multi output
    if len(y_pred.shape) == 2:
        df = pd.DataFrame()
        for j in range(y_pred.shape[1]):
            df["y_test_output_{}".format(j)] = y_test[:, j]
            df["y_pred_output_{}".format(j)] = y_pred[:, j]
        run.log_dataframe(df=df.head(nrows), name="test_predictions")


def _log_test_preds_proba(run, classifier, X_test, nrows=1000):
    try:
        import pandas as pd
    except ImportError:
        return

    try:
        y_pred_proba = classifier.predict_proba(X_test)
    except Exception as e:
        print(
            "This classifier does not provide predictions probabilities. Error: {}".format(
                e
            )
        )
        return

    df = pd.DataFrame(data=y_pred_proba, columns=classifier.classes_)
    run.log_dataframe(df=df.head(nrows), name="test_proba_predictions")


def log_regressor(regressor, X_test, y_test, nrows=1000, run=None):
    assert is_regressor(regressor), "regressor should be sklearn regressor."

    run = tracking.get_or_create_run(run)

    run.log_inputs(**regressor.get_params())

    y_pred = regressor.predict(X_test)

    # single output
    results = {}
    if len(y_pred.shape) == 1:
        results["evs"] = explained_variance_score(y_test, y_pred)
        results["me"] = max_error(y_test, y_pred)
        results["mae"] = mean_absolute_error(y_test, y_pred)
        results["r2"] = r2_score(y_test, y_pred)
    # multi output
    if len(y_pred.shape) == 2:
        results["r2"] = regressor.score(X_test, y_test)
    run.log_metrics(**results)

    _log_test_predictions(run, y_test, y_pred=y_pred, nrows=nrows)


def log_classifier(classifier, X_test, y_test, nrows=1000, run=None):
    assert is_classifier(classifier), "classifier should be sklearn classifier."

    run = tracking.get_or_create_run(run)

    run.log_inputs(**sanitize_dict(classifier.get_params()))

    _log_test_preds_proba(run, classifier, X_test, nrows=nrows)

    y_pred = classifier.predict(X_test)

    results = {}
    for metric_name, values in zip(
        ["precision", "recall", "fbeta_score", "support"],
        precision_recall_fscore_support(y_test, y_pred),
    ):
        for i, value in enumerate(values):
            results["{}_class_{}_test".format(metric_name, i)] = value
    results["accuracy"] = accuracy_score(y_test, y_pred)
    results["f1"] = f1_score(y_pred, y_pred, average="weighted")
    run.log_metrics(**results)
    _log_test_predictions(run, y_test, y_pred=y_pred, nrows=nrows)



from clipped.utils.paths import module_type

from polyaxon._constants.globals import UNKNOWN
from traceml.events import V1EventChart, V1EventChartKind
from traceml.logger import logger
from traceml.processors.errors import (
    BOKEH_ERROR_MESSAGE,
    MATPLOTLIB_ERROR_MESSAGE,
    PLOTLY_ERROR_MESSAGE,
)

try:
    import numpy as np
except ImportError:
    np = None


def bokeh_chart(figure) -> V1EventChart:
    try:
        from bokeh.embed import json_item
    except ImportError:
        logger.warning(BOKEH_ERROR_MESSAGE)
        return UNKNOWN
    return V1EventChart(kind=V1EventChartKind.BOKEH, figure=json_item(figure))


def altair_chart(figure) -> V1EventChart:
    return V1EventChart(kind=V1EventChartKind.VEGA, figure=figure.to_dict())


def plotly_chart(figure) -> V1EventChart:
    if module_type(figure, "matplotlib.figure.Figure"):
        try:
            from traceml.vendor.matplotlylib import mpl_to_plotly
        except ImportError:
            logger.error(MATPLOTLIB_ERROR_MESSAGE)
            logger.error(PLOTLY_ERROR_MESSAGE)
            return UNKNOWN

        figure = mpl_to_plotly(figure)
    else:
        try:
            import plotly.tools
        except ImportError:
            logger.error(PLOTLY_ERROR_MESSAGE)
            return UNKNOWN

        figure = plotly.tools.return_figure_from_figure_or_data(
            figure, validate_figure=True
        )
    return V1EventChart(kind=V1EventChartKind.PLOTLY, figure=figure)


def mpl_plotly_chart(figure, close: bool = True) -> V1EventChart:
    try:
        import plotly.tools

        from plotly import optional_imports
    except ImportError:
        logger.warning(PLOTLY_ERROR_MESSAGE)
        return UNKNOWN

    try:
        import matplotlib
        import matplotlib.pyplot as plt

        from matplotlib.figure import Figure
    except ImportError:
        logger.warning(MATPLOTLIB_ERROR_MESSAGE)

    if module_type(figure, "matplotlib.figure.Figure"):
        pass
    else:
        if figure == matplotlib.pyplot:
            figure = figure.gcf()
        elif not isinstance(figure, Figure):
            if hasattr(figure, "figure"):
                figure = figure.figure
                # Some matplotlib objects have a figure function
                if not isinstance(figure, Figure):
                    raise ValueError(
                        "Only matplotlib.pyplot or matplotlib.pyplot.Figure objects are accepted."
                    )

    from traceml.vendor.matplotlylib import mpl_to_plotly

    plotly_figure = mpl_to_plotly(figure)
    result = plotly_chart(figure=plotly_figure)
    if close:
        try:
            plt.close(figure.number)
        except Exception:  # noqa
            plt.close(figure)

    return result



from typing import Dict, List

from clipped.utils.np import to_np

from polyaxon._constants.globals import UNKNOWN
from traceml.artifacts import V1ArtifactKind
from traceml.events import (
    LoggedEventSpec,
    V1Event,
    V1EventConfusionMatrix,
    V1EventCurve,
    V1EventCurveKind,
    V1EventHistogram,
)
from traceml.logger import logger
from traceml.processors.errors import NUMPY_ERROR_MESSAGE, SKLEARN_ERROR_MESSAGE

try:
    import numpy as np
except ImportError:
    np = None


def metric(value):
    if isinstance(value, float):
        return value

    if not np:
        logger.warning(NUMPY_ERROR_MESSAGE)
        return UNKNOWN

    value = to_np(value)
    assert value.squeeze().ndim == 0, "scalar should be 0D"
    return float(value)


def histogram(values, bins, max_bins=None):
    if not np:
        logger.warning(NUMPY_ERROR_MESSAGE)
        return UNKNOWN

    values = to_np(values).astype(float)

    if values.size == 0:
        raise ValueError("The input has no element.")
    values = values.reshape(-1)
    values, counts = np.histogram(values, bins=bins)

    if counts.size == 0:
        logger.warning("Tracking an empty histogram")
        return UNKNOWN

    return np_histogram(values=values, counts=counts, max_bins=max_bins)


def np_histogram(values, counts, max_bins=None):
    try:
        values = values.tolist()
        counts = counts.tolist()
    except:  # noqa
        pass
    max_bins = max_bins or 512
    values_len = len(values)
    counts_len = len(counts)
    if values_len > max_bins:
        raise ValueError(
            "The maximum bins for a histogram is {}, received {}".format(
                max_bins, values_len
            )
        )
    if values_len + 1 != counts_len:
        raise ValueError("len(hist.values) must be len(hist.counts) + 1")
    return V1EventHistogram(values=values, counts=counts)


def roc_auc_curve(fpr, tpr, auc=None):
    return V1EventCurve(
        kind=V1EventCurveKind.ROC,
        x=fpr,
        y=tpr,
        annotation=str(auc) if auc else None,
    )


def sklearn_roc_auc_curve(y_preds, y_targets, pos_label=None):
    try:
        from sklearn.metrics import auc, roc_curve
    except ImportError:
        logger.warning(SKLEARN_ERROR_MESSAGE)

    try:
        y_true = y_targets.numpy()
    except AttributeError:
        y_true = y_targets
    try:
        y_pred = y_preds.numpy()
    except AttributeError:
        y_pred = y_preds
    fpr, tpr, _ = roc_curve(y_true, y_pred, pos_label=pos_label)
    auc_score = auc(fpr, tpr)
    return V1EventCurve(
        kind=V1EventCurveKind.ROC,
        x=fpr,
        y=tpr,
        annotation=str(auc_score),
    )


def pr_curve(precision, recall, average_precision=None):
    return V1EventCurve(
        kind=V1EventCurveKind.PR,
        x=precision,
        y=recall,
        annotation=str(average_precision) if average_precision else None,
    )


def sklearn_pr_curve(y_preds, y_targets, pos_label=None):
    try:
        from sklearn.metrics import average_precision_score, precision_recall_curve
    except ImportError:
        logger.warning(SKLEARN_ERROR_MESSAGE)

    try:
        y_true = y_targets.numpy()
    except AttributeError:
        y_true = y_targets
    try:
        y_pred = y_preds.numpy()
    except AttributeError:
        y_pred = y_preds

    precision, recall, _ = precision_recall_curve(y_true, y_pred, pos_label=pos_label)
    ap = average_precision_score(y_true, y_pred)
    return V1EventCurve(
        kind=V1EventCurveKind.PR,
        x=precision,
        y=recall,
        annotation=str(ap),
    )


def curve(x, y, annotation=None):
    return V1EventCurve(
        kind=V1EventCurveKind.CUSTOM,
        x=x,
        y=y,
        annotation=str(annotation) if annotation else None,
    )


def confusion_matrix(x, y, z):
    if hasattr(x, "tolist"):
        x = x.tolist()
    if hasattr(x, "tolist"):
        y = y.tolist()
    if hasattr(x, "tolist"):
        z = z.tolist()
    try:
        x_len = len(x)
        y_len = len(y)
        z_len = len(z)
        if x_len != y_len or x_len != z_len:
            raise ValueError(
                "Received invalid data for confusion matrix. "
                "All arrays must have the same structure: "
                "[len(x): {}, len(y): {}, len(z): {}]".format(
                    x_len,
                    y_len,
                    z_len,
                )
            )
        zi_len = [len(zi) for zi in z]
        if len(set(zi_len)) != 1 or zi_len[0] != z_len:
            raise ValueError(
                "Received invalid data for confusion matrix. "
                "Current structure: [len(x): {}, len(y): {}, len(z): {}]. "
                "The z array has different nested structures: {}".format(
                    x_len, y_len, z_len, zi_len
                )
            )
    except Exception as e:  # noqa
        raise ValueError(
            "Received invalid data for confusion matrix. Error {}".format(e)
        )
    return V1EventConfusionMatrix(
        x=x,
        y=y,
        z=z,
    )


def metrics_dict_to_list(metrics: Dict) -> List:
    results = []
    for k, v in metrics.items():
        results.append(
            LoggedEventSpec(
                name=k,
                kind=V1ArtifactKind.METRIC,
                event=V1Event.make(metric=v),
            )
        )
    return results
